import requests
import time
import pandas as pd

from .utils import get_causadb_url


class Model:
    def __init__(self, model_name: str, client: "CausaDB") -> None:
        """Initializes the Model class.

        Args:
            model_name (str): The name of the model.
            client (CausaDB): A CausaDB client.
        """
        self.client = client
        self.model_name = model_name
        self.config = {}

        # Pull config from the server
        headers = {"token": self.client.token_secret}
        response = requests.get(
            f"{get_causadb_url()}/models/{self.model_name}",
            headers=headers,
        ).json()

        if "details" in response:
            self.config = response["details"]["config"]

        self._update()

    def __repr__(self) -> str:
        return f"<Model {self.model_name}>"

    def remove(self) -> None:
        """Remove the model from the CausaDB system."""
        headers = {"token": self.client.token_secret}
        try:
            requests.delete(
                f"{get_causadb_url()}/models/{self.model_name}",
                headers=headers,
            )
        except:
            raise Exception("CausaDB server request failed")

    def set_nodes(self, nodes: list[str]) -> None:
        """Set the nodes of the model.

        Args:
            nodes (list[str]): A list of node names.

        Example:
            >>> model.set_nodes(["x", "y", "z"])
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models/{self.model_name}", headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        self.config = response["details"]["config"]
        self.config["nodes"] = nodes

        self._update()

    def get_nodes(self) -> list[str]:
        """Get the nodes of the model.

        Returns:
            list[str]: A list of node names.
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models/{self.model_name}", headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        return response["details"]["config"]["nodes"]

    def set_edges(self, edges: list[tuple[str, str]]) -> None:
        """Set the edges of the model.

        Args:
            edges (list[tuple[str, str]]): A list of tuples representing edges.

        Example:
            >>> model.set_edges([
            ...     ("SaturatedFatsInDiet", "Weight"),
            ...     ("Weight", "BMI"),
            ... ])
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models/{self.model_name}", headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        self.config = response["details"]["config"]
        self.config["edges"] = edges

        self._update()

    def get_edges(self) -> list[tuple[str, str]]:
        """Get the edges of the model.

        Returns:
            list[tuple[str, str]]: A list of tuples representing edges.
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models/{self.model_name}", headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        edges = response["details"]["config"]["edges"]
        # Convert the edges to a list of tuples
        return [(edge[0], edge[1]) for edge in edges]

    def set_node_types(self, node_types: dict) -> None:
        """Set the node types of the model.

        Args:
            node_types (dict): A dictionary of node types.

        Example:
            >>> model.set_node_types({
            ...     "x1": {"type": "seasonal", "min": 0, "max": 1}
            ... })
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models/{self.model_name}", headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        self.config = response["details"]["config"]
        self.config["node_types"] = node_types

        self._update()

    def get_node_types(self) -> dict:
        """Get the node types of the model.

        Returns:
            dict: A dictionary of node types.
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models/{self.model_name}", headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        return response["details"]["config"]["node_types"]

    def attach(self, data_name: str) -> None:
        """Attach data to the model.

        Args:
            data_name (str): The name of the data to attach.
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.post(
                f"{get_causadb_url()}/models/{self.model_name}/attach/{data_name}",
                headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

    def detach(self, data_name: str) -> None:
        """Detach data from the model.

        Args:
            data_name (str): The name of the data to detach.
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.delete(
                f"{get_causadb_url()}/models/{self.model_name}/detach",
                headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

    def train(self, wait=True, poll_interval=0.2) -> None:
        """Train the model.

        Args:
            wait (bool): Whether to wait for the model to finish training.
            poll_interval (float): The interval at which to poll the server for the model status.

        Example:
            >>> model.train()
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.post(
                f"{get_causadb_url()}/models/{self.model_name}/train",
                headers=headers
            )
        except:
            raise Exception("CausaDB server request failed")

        # If HTTPException status code is 400, raise an exception
        if response.status_code == 400:
            raise Exception(response.json()["detail"])
        response = response.json()

        if wait:
            while self.status() != "trained":
                # Try again in 200ms
                time.sleep(poll_interval)

    def status(self) -> str:
        """Get the status of the model.

        Returns:
            str: The status of the model.
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models/{self.model_name}",
                headers=headers,
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        model_status = response["details"]["status"]

        return model_status

    def simulate_actions(self, action: dict) -> dict:
        """Simulate an action on the model.

        Args:
            action (dict): A dictionary representing the action.

        Returns:
            dict: A dictionary representing the result of the action.
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.post(
                f"{get_causadb_url()}/models/{self.model_name}/simulate-actions",
                headers=headers,
                json=action,
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        if "outcome" in response:
            return response["outcome"]

        raise Exception("CausaDB server request failed")

    def optimal_actions(self, target_outcomes: dict[str, float], actionable_nodes: list[str], condition_nodes: dict[str, float] = {}) -> dict:
        """Get the optimal actions for a given set of target outcomes.

        Args:
            target_outcomes (dict[str, float]): A dictionary of target outcomes.
            actionable_nodes (list[str]): A list of actionable nodes.
            condition_nodes (dict[str, float]): A dictionary of condition nodes.

        Returns:
            dict: A dictionary representing the optimal actions.

        Example:
            >>> model.optimal_actions(
            ...     {"x": 0.5},
            ...     ["y"],
            ...     {"z": 0.5}
            ... )
        """
        headers = {"token": self.client.token_secret}

        try:
            response = requests.post(
                f"{get_causadb_url()}/models/{self.model_name}/optimal-actions",
                headers=headers,
                json={
                    "target_outcomes": target_outcomes,
                    "actionable_nodes": actionable_nodes,
                    "condition_nodes": condition_nodes
                },
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        if "optimal_actions" in response:
            return response["optimal_actions"]

        raise Exception("CausaDB server request failed")

    def _update(self) -> None:
        """Pushes the current state of the model to the CausaDB server."""
        headers = {"token": self.client.token_secret}

        try:
            response = requests.post(
                f"{get_causadb_url()}/models/{self.model_name}",
                headers=headers,
                json=self.config
            ).json()
        except:
            raise Exception("CausaDB server request failed")



import typer
import requests
from causadb.cli.utils import load_config, show_table, CAUSADB_URL
from typing import Annotated
import json

app = typer.Typer()


@app.command()
def list():
    """
    List models.
    """
    config = load_config()
    token_secret = config["default"]["token_secret"]

    headers = {"token": token_secret}

    data = requests.get(
        f"{CAUSADB_URL}/models", headers=headers
    ).json()

    show_table(data["models"], columns=[
               "id", "name", "data", "created_at"])


@app.command()
def add(
    model_name_raw: Annotated[str, typer.Option(
        "--model",
        help="The name of your model, used for accessing it later.")] = None,
    model_config_filepath: Annotated[str, typer.Option(
        "--config",
        help="Path to a model config file (e.g. /path/to/config.json). See the docs for more information.")] = None
):
    """
    Create a model.
    """

    # If name is None, prompt the user for a name.
    if model_name_raw is None:
        model_name_raw = typer.prompt(
            "Enter the name of your model (e.g. my-model)")

    model_name = model_name_raw.replace(" ", "-")

    # Replace spaces with dashes.
    if " " in model_name_raw:
        typer.echo(
            f"Note: Spaces in model name replaced with dashes. Now model name: {model_name}.")

    # If model_config_filepath is None, prompt the user for a model_config_filepath file.
    if model_config_filepath is None:
        model_config_filepath = typer.prompt(
            "Enter the path to your model config file (e.g. /path/to/config.json)")

    # Read the config file into a dict.
    try:
        with open(model_config_filepath, "r") as f:
            try:
                model_config = json.load(f)
            except:
                typer.echo(
                    "Failed to read config file. Please check the file is valid JSON.")
                return 1
    except:
        typer.echo("Failed to read config file. Please check the path.")
        return 1

    config = load_config()
    token_secret = config["default"]["token_secret"]

    headers = {"token": token_secret}

    data = requests.post(
        f"{CAUSADB_URL}/models/{model_name}",
        headers=headers,
        json=model_config,
    ).json()

    typer.echo(data["message"])


@app.command()
def remove(
    model_name: Annotated[str, typer.Option(
        "--model",
        help="The name of the model you wish to remove.")] = None,
):
    """
    Delete a model.
    """
    # If name is None, prompt the user for a name.
    if model_name is None:
        model_name = typer.prompt(
            "Enter the name of the model you wish to remove (e.g. my-model)")

    # Make sure model_name can be used in an API URL.
    if " " in model_name:
        typer.echo(
            "Model name cannot contain spaces. Please try again.")
        return 1

    # Load the configuration and get the token secret
    config = load_config()
    token_secret = config["default"]["token_secret"]

    headers = {"token": token_secret}

    # Make a request to the remove data endpoint
    response = requests.delete(
        f"{CAUSADB_URL}/models/{model_name}",
        headers=headers
    ).json()

    typer.echo(response["message"])


@app.command()
def info(
    model_name: Annotated[str, typer.Option(
        "--model",
        help="The name of the model you wish to retrieve information about.")] = None
):
    """
    Show info about a model.
    """

    if model_name is None:
        model_name = typer.prompt(
            "Enter the name of the model you wish to attach data to (e.g. my-model)")

    config = load_config()
    token_secret = config["default"]["token_secret"]

    headers = {"token": token_secret}

    data = requests.get(
        f"{CAUSADB_URL}/models/{model_name}",
        headers=headers,
    ).json()

    # Pretty print the response
    typer.echo(json.dumps(data, indent=4))


@app.command()
def attach(
    model_name: Annotated[str, typer.Option(
        "--model",
        help="The name of the model you wish to attach data to.")] = None,
    data_name: Annotated[str, typer.Option(
        "--data",
        help="The name of the data you wish to attach.")] = None,
):
    """
    Attach data to a model.
    """

    if model_name is None:
        model_name = typer.prompt(
            "Enter the name of the model you wish to attach data to (e.g. my-model)")

    if data_name is None:
        data_name = typer.prompt(
            "Enter the name of the data you wish to attach (e.g. my-data)")

    config = load_config()
    token_secret = config["default"]["token_secret"]

    headers = {"token": token_secret}

    data = requests.post(
        f"{CAUSADB_URL}/models/{model_name}/attach/{data_name}",
        headers=headers,
    ).json()

    typer.echo(f"{data_name} successfully attached to {model_name}.")


@app.command()
def detach(
    model_name: Annotated[str, typer.Option(
        "--model",
        help="The name of the model you wish to attach data to.")] = None
):
    """
    Detach a datasource from a model.
    """

    if model_name is None:
        model_name = typer.prompt(
            "Enter the name of the model you wish to attach data to (e.g. my-model)")

    config = load_config()
    token_secret = config["default"]["token_secret"]

    headers = {"token": token_secret}

    data = requests.delete(
        f"{CAUSADB_URL}/models/{model_name}/detach",
        headers=headers,
    ).json()

    typer.echo("Data successfully detached from model.")


@app.command()
def train(
    model_name: Annotated[str, typer.Option(
        "--model",
        help="The name of the model you wish to train.")] = None
):
    """
    Train a model.
    """

    if model_name is None:
        model_name = typer.prompt(
            "Enter the name of the model you wish to train (e.g. my-model)")

    config = load_config()
    token_secret = config["default"]["token_secret"]

    headers = {"token": token_secret}

    data = requests.post(
        f"{CAUSADB_URL}/models/{model_name}/train",
        headers=headers,
    ).json()

    if data["status"] == "success":
        typer.echo(
            f"Model training started. Check the status with `causadb models info --model {model_name}`.")
    else:
        typer.echo("Model training failed. Check the logs for more information.")


@app.command()
def status(
    model_name: Annotated[str, typer.Option(
        "--model",
        help="The name of the model you wish to retrieve status for.")] = None
):
    """
    Show status of a model.
    """

    if model_name is None:
        model_name = typer.prompt(
            "Enter the name of the model you wish to attach data to (e.g. my-model)")

    config = load_config()
    token_secret = config["default"]["token_secret"]

    headers = {"token": token_secret}

    data = requests.get(
        f"{CAUSADB_URL}/models/{model_name}",
        headers=headers,
    ).json()

    model_status = data["details"]["status"]

    typer.echo(f"Model status: {model_status}")


@app.command()
def simulate_actions(
    model_name: Annotated[str, typer.Option(
        "--model",
        help="The name of the model you wish to train.")] = None,
    action_config_filepath: Annotated[str, typer.Option(
        "--action",
        help="Path to the action config (e.g. /path/to/action.json)")] = None
):
    """
    Simulate an action.
    """

    if model_name is None:
        model_name = typer.prompt(
            "Enter the name of the model you wish to train (e.g. my-model)")

    if action_config_filepath is None:
        action_config_filepath = typer.prompt(
            "Enter the path to your action config file (e.g. /path/to/config.json)")

    config = load_config()
    token_secret = config["default"]["token_secret"]

    headers = {"token": token_secret}

    action_config = json.load(open(action_config_filepath, "r"))

    data = requests.post(
        f"{CAUSADB_URL}/models/{model_name}/simulate-actions",
        headers=headers,
        json=action_config,
    ).json()

    if data["status"] == "success":
        typer.echo(json.dumps(data["outcome"], indent=4))



import requests
import os
from .data import Data
from .model import Model
from .utils import get_causadb_url, set_causadb_url


class CausaDB:
    """CausaDB client class to interact with the CausaDB system.
    """

    def __repr__(self) -> str:
        return f"<CausaDB client>"

    def __str__(self) -> str:
        return f"{self.token_id}"

    def __init__(self, custom_url=None) -> None:
        """Initializes the CausaDB client.

        Args:
            custom_url (str, optional): The URL of the CausaDB server. For custom deployments or development purposes. Defaults to None.
        """
        self.token_id = None
        self.token_secret = None
        if custom_url is not None:
            set_causadb_url(custom_url)

    def set_token(self, token_id: str, token_secret: str) -> bool:
        """Set the token for the CausaDB client.

        Args:
            token_id (str): Token ID provided by CausaDB.
            token_secret (str): Token secret provided by CausaDB.

        Returns:
            bool: True if the token is valid, False otherwise.
        """

        # Verify that the tokens are correct
        headers = {"token": token_secret}
        response = requests.get(
            f"{get_causadb_url()}/account",
            headers=headers
        )

        # If the response is successful, set the tokens
        if response.status_code == 200:
            self.token_id = token_id
            self.token_secret = token_secret
        else:
            raise Exception("Invalid token")

    def create_model(self, model_name: str) -> Model:
        """Create a model and add it to the CausaDB system.

        Args:
            model_name (str): The name of the model.

        Returns:
            Model: The model object.
        """
        return Model(model_name, self)

    def add_data(self, data_name: str) -> Data:
        """Add data to the CausaDB system.

        Args:
            data_name (str): The name of the data.

        Returns:
            Data: The data object.
        """
        return Data(data_name, self)

    def get_model(self, model_name: str) -> Model:
        """Get a model by name.

        Args:
            model_name (str): The name of the model.

        Returns:
            Model: The model object.
        """
        headers = {"token": self.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models/{model_name}", headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        # If the model exists, return it
        model = Model(model_name, self)

        return model

    def list_models(self) -> list[Model]:
        """List all models.

        Returns:
            list[Model]: A list of model objects.
        """
        headers = {"token": self.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models", headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        model_list = []
        for model_spec in response.get("models", []):
            model = Model(model_spec["name"], self)
            model_list.append(model)

        return model_list

    def get_data(self, data_name: str) -> Data:
        """Get a data by name.

        Args:
            data_name (str): The name of the data.

        Returns:
            Data: The data object.
        """
        headers = {"token": self.token_secret}
        try:
            response = requests.get(
                f"{get_causadb_url()}/data/{data_name}", headers=headers
            ).json()
        except:
            raise Exception("CausaDB server request failed")

        data = Data(data_name, self)

        return data

    def list_data(self) -> list[Data]:
        """List all data.

        Returns:
            list[Data]: A list of data objects.
        """
        headers = {"token": self.token_secret}

        try:
            response = requests.get(
                f"{get_causadb_url()}/data", headers=headers
            ).json()
        except:
            raise Exception("CausaDB client failed to connect to server")

        data_list = []
        for data_spec in response.get("data", []):
            data = Data(data_spec["name"], self)
            data_list.append(data)

        return data_list



import os
import dotenv
import toml
from rich.console import Console
from rich.table import Table
dotenv.load_dotenv()


CAUSADB_URL = os.getenv("CAUSADB_URL", "https://api.causadb.com/v1")


def load_config():
    # Get config from ~/.causadb/config.toml
    dir_name = os.path.expanduser("~/.causadb")
    config_filepath = os.path.join(dir_name, "config.toml")

    if not os.path.exists(config_filepath):
        return None

    with open(config_filepath, "r") as f:
        config = toml.load(f)

    return config


def show_table(data, columns=None):
    table = Table(show_header=True, header_style="bold turquoise4")

    if columns is None:
        columns = list(data[0].keys())

    for column in columns:
        table.add_column(column)

    for item in data:
        table.add_row(*[str(item[column]) for column in columns])

    console = Console()
    console.print(table)
